﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using LockCrypt.Core;

namespace LockCrypt.DAL {
    /// <summary>
    /// Enables read/write access to a database file.
    /// </summary>
    public abstract class FileBasedDatabaseProvider : FileBasedDatabase, IDbProvider {
        /// <summary>
        /// Gets the encryption provider.
        /// </summary>
        /// <value>The encryption provider.</value>
        public abstract IEncryptionService Encryption { get; }

        /// <summary>
        /// Gets the type of database which this provider can access.
        /// </summary>
        /// <value>The type of database which this provider can access.</value>
        public abstract Constants.DatabaseType DatabaseType { get; }

        /// <summary>
        /// Gets a formatted description of this database connection, either the current host or file.
        /// </summary>
        /// <value>A formatted description of this database connection.</value>
        public abstract string FormattedName { get; }

        /// <summary>
        /// Gets a value indicating whether this instance is either connected or can connect.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance is connected or can connect, otherwise <c>false</c>.
        /// </value>
        public abstract bool IsConnected { get; }

        /// <summary>
        /// Gets whether an attempt can be made to open a database connection.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if an attempt can be made to open a database connection, otherwise <c>false</c>.
        /// </value>
        public abstract bool CanAttemptConnection { get; }

        /// <summary>
        /// Gets the database connection string.
        /// </summary>
        /// <value>The connection string.</value>
        public abstract string ConnectionString { get; }

        /// <summary>
        /// Initializes a new <see cref="FileBasedDatabase"/>.
        /// </summary>
        /// <param name="fileName">Name of the database file.</param>
        protected FileBasedDatabaseProvider(string fileName) 
            :base(fileName) {
        }

        /// <summary>
        /// Checks the password entered is valid.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if the current password is valid, otherwise <c>false</c>.
        /// </returns>
        public abstract bool CheckPassword();

        /// <summary>
        /// Adds accounts to the default database.
        /// </summary>
        /// <param name="accounts">The accounts to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        public abstract void AddAccounts(IEnumerable<IAccount> accounts, ILogProvider logger);

        /// <summary>
        /// Edits an account.
        /// </summary>
        /// <param name="account">The account.</param>
        public abstract void EditAccount(IAccount account);

        /// <summary>
        /// Deletes an account.
        /// </summary>
        /// <param name="accountId">The account ID.</param>
        public abstract void DeleteAccount(string accountId);

        /// <summary>
        /// Deletes multiple accounts.
        /// </summary>
        /// <param name="accountIds">The account IDs.</param>
        public abstract void DeleteAccounts(IEnumerable<string> accountIds);

        /// <summary>
        /// Gets a list of recipes from the current database file.
        /// </summary>
        /// <returns>A list of recipes in the database.</returns>
        public abstract List<IAccount> GetAccounts();

        /// <summary>
        /// Adds account groups to the default database.
        /// </summary>
        /// <param name="groups">The groups to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        public abstract void AddGroups(IEnumerable<IGroup> groups, ILogProvider logger);

        /// <summary>
        /// Edits a group.
        /// </summary>
        /// <param name="group">The group.</param>
        public abstract void EditGroup(IGroup group);

        /// <summary>
        /// Deletes a group and all child groups/accounts.
        /// </summary>
        /// <param name="group">The group.</param>
        public abstract void DeleteGroup(IGroup group);

        /// <summary>
        /// Deletes multiple groups and their child accounts. Does not delete child groups.
        /// </summary>
        /// <param name="groupIds">The group IDs.</param>
        public abstract void DeleteGroups(IEnumerable<string> groupIds);

        /// <summary>
        /// Gets a list of account groups from the current database.
        /// </summary>
        /// <returns>A list of account groups in the database.</returns>
        public abstract List<IGroup> GetGroups();

        /// <summary>
        /// Adds templates to the default database.
        /// </summary>
        /// <param name="templates">The templates to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        /// <returns><c>null</c> on success, a log of errors on failure.</returns>
        public abstract void AddTemplates(IEnumerable<ITemplate> templates, ILogProvider logger);

        /// <summary>
        /// Edits a template.
        /// </summary>
        /// <param name="template">The template.</param>
        public abstract void EditTemplate(ITemplate template);

        /// <summary>
        /// Deletes a template.
        /// </summary>
        /// <param name="templateId">The template ID.</param>
        public abstract void DeleteTemplate(string templateId);

        /// <summary>
        /// Gets a list of templates from the current database file.
        /// </summary>
        /// <returns>A list of templates in the database.</returns>
        public abstract List<ITemplate> GetTemplates();

        /// <summary>
        /// Gets matching templates by name.
        /// </summary>
        /// <param name="templateName">Name of the template.</param>
        /// <returns>All templates with a matching name.</returns>
        public abstract List<ITemplate> GetTemplatesByName(string templateName);

        /// <summary>
        /// Creates a database and required tables.
        /// </summary>
        /// <exception cref="System.Exception">Multiple exceptions, check the reference for DirectoryInfo.Create and the FileInfo constructor.</exception>
        public abstract void CreateDatabase();

        /// <summary>
        /// Clears the database.
        /// </summary>
        public abstract void ClearDatabase();

        /// <summary>
        /// Imports a database without clearing existing data.
        /// </summary>
        /// <param name="contents">The contents to import.</param>
        /// <param name="logger">The logger.</param>
        public abstract void ImportDatabase(DatabaseContents contents, ILogProvider logger);

        /// <summary>
        /// Changes the encryption password.
        /// </summary>
        /// <param name="newPassword">The new password.</param>
        public abstract void ChangePassword(string newPassword);
    }
}
