﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SQLite;
using System.Linq;
using System.Text;
using Lime49.Exceptions;
using LockCrypt.Core;
using System.IO;
using System.Collections;
using System.Text.RegularExpressions;

namespace LockCrypt.DAL {
    //public class SQLiteProvider : FileBasedDatabase, IDbProvider {
    public class SQLiteProvider : FileBasedDatabaseProvider {
        private readonly IEncryptionService _encryptionService;
        private bool? _isConnected;
        private Version _dbVersion;

        /// <summary>
        /// Gets the encryption provider.
        /// </summary>
        /// <value>The encryption provider.</value>
        public override IEncryptionService Encryption {
            get { return _encryptionService; }
        }

        /// <summary>
        /// Gets the type of database which this provider can access.
        /// </summary>
        /// <value>The type of database which this provider can access.</value>
        public override Constants.DatabaseType DatabaseType {
            get { return Constants.DatabaseType.SQLite; }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is either connected or can connect.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance is connected or can connect, otherwise <c>false</c>.
        /// </value>
        public override bool IsConnected {
            get {
                if(!_isConnected.HasValue) {
                    _isConnected = CheckPassword(true);
                }
                return _isConnected.Value;
            }
        }

        /// <summary>
        /// Gets whether an attempt can be made to open a database connection.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if an attempt can be made to open a database connection, otherwise <c>false</c>.
        /// </value>
        public override bool CanAttemptConnection {
            get { return File.Exists(FileName); }
        }

        /// <summary>
        /// Gets a formatted description of this database connection.
        /// </summary>
        /// <value>A formatted description of this database connection.</value>
        public override string FormattedName {
            get {
                try {
                    return string.Format("sqlite://{0}", Path.GetFullPath(FileName));
                } catch {
                    return string.Format("sqlite://{0}", FileName);
                }
            }
        }

        /// <summary>
        /// The connection string for SQLite. Placeholder1 = filename
        /// </summary>
        public const string SQLiteConnectionString = "Data Source={0}";

        /// <summary>
        /// Gets the database connection string.
        /// </summary>
        /// <value>The connection string.</value>
        public override string ConnectionString {
            get { return string.Format(SQLiteConnectionString, FileName); }
        }

        /// <summary>
        /// Initializes a new <see cref="SQLiteProvider"/>.
        /// </summary>
        /// <param name="pathOrConnectionString">Either an SQLite connection string or a file path.</param>
        public SQLiteProvider(string pathOrConnectionString)
            :base(pathOrConnectionString) {
            Match match = Regex.Match(pathOrConnectionString, @"\s*?Data Source=(.*)", RegexOptions.IgnoreCase | RegexOptions.CultureInvariant);
            if(match.Success) {
                // if the connection string matches, replace update the path with the matched connection string path
                this.FileName = match.Groups[1].Value;
            }
            _encryptionService = new NullEncryptionService();

        }

        /// <summary>
        /// Initializes a new <see cref="SQLiteProvider"/> for a specified file.
        /// </summary>
        /// <param name="pathOrConnectionString">Either an SQLite connection string or a file path.</param>
        /// <param name="encryptionService">The encryption service.</param>
        public SQLiteProvider(string pathOrConnectionString, IEncryptionService encryptionService)
            : this(pathOrConnectionString) {
            _encryptionService = encryptionService;
        }

        #region Accounts
        /// <summary>
        /// Adds accounts to the default database.
        /// </summary>
        /// <param name="accounts">The accounts to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        /// <returns><c>null</c> on success, a log of errors on failure.</returns>
        public override void AddAccounts(IEnumerable<IAccount> accounts, ILogProvider logger) {
            SQLiteConnection conn = null;
            if(logger == null) {
                // if there's no log delegate, log to console.
                logger = _logger;
            }
            MultipleExceptionLogger logMultiple = new MultipleExceptionLogger();
            CompositeLogger logComposite = new CompositeLogger(logger, logMultiple);
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    AddAccounts(conn, trans, accounts, logComposite);
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
            if(logMultiple.MessageLog.Count > 0) {
                logMultiple.MessageLog.Insert(0, new Exception(I18nUtils.GetString("Errors", "NonFatalErrors")));
                throw new MultipleException(logMultiple.MessageLog);
            }
        }

        /// <summary>
        /// Add accounts using an existing transation.
        /// </summary>
        /// <param name="conn">The connection.</param>
        /// <param name="transaction">The transaction.</param>
        /// <param name="accounts">The accounts.</param>
        /// <param name="logger">The logger.</param>
        private void AddAccounts(SQLiteConnection conn, SQLiteTransaction transaction, IEnumerable<IAccount> accounts, ILogProvider logger) {
            using(SQLiteCommand cmdAccount = conn.CreateCommand())
            using(SQLiteCommand cmdAddFields = conn.CreateCommand()) {
                cmdAccount.Transaction = transaction;
                cmdAddFields.Transaction = transaction;
                cmdAccount.CommandText = "INSERT INTO accounts (accountid, name, parentid, templateid, created, modified, icon, notes, autotype) VALUES (@accountid, @name, @parentid, @templateid, @created, @modified, @icon, @notes, @autotype)";
                cmdAccount.Parameters.Add("@accountid", DbType.Binary);
                cmdAccount.Parameters.Add("@name", DbType.Binary);
                cmdAccount.Parameters.Add("@parentid", DbType.Binary);
                cmdAccount.Parameters.Add("@templateid", DbType.Binary);
                cmdAccount.Parameters.AddWithValue("@created", DateTime.Now);
                cmdAccount.Parameters.AddWithValue("@modified", DateTime.Now);
                cmdAccount.Parameters.Add("@icon", DbType.Binary);
                cmdAccount.Parameters.Add("@notes", DbType.Binary);
                cmdAccount.Parameters.Add("@autotype", DbType.Binary);
                cmdAddFields.CommandText = "INSERT INTO accountfields (accountid, fieldname, fieldvalue, meta, sortindex) VALUES (@accountid, @fieldname, @fieldvalue, @meta, @sortindex)";
                cmdAddFields.Parameters.Add("@accountid", DbType.Binary);
                cmdAddFields.Parameters.Add("@fieldname", DbType.Binary);
                cmdAddFields.Parameters.Add("@fieldvalue", DbType.Binary);
                cmdAddFields.Parameters.Add("@meta", DbType.Binary);
                cmdAddFields.Parameters.Add("@sortindex", DbType.Int32);
                foreach(IAccount account in accounts) {
                    try {
                        cmdAccount.Parameters["@accountid"].Value = _encryptionService.EncryptString(account.ID);
                        cmdAccount.Parameters["@name"].Value = _encryptionService.EncryptString(account.Name);
                        cmdAccount.Parameters["@parentid"].Value = _encryptionService.EncryptString(account.ParentID);
                        cmdAccount.Parameters["@templateid"].Value = account.Template == null ? null : _encryptionService.EncryptString(account.Template.ID);
                        cmdAccount.Parameters["@created"].Value = account.Created;
                        cmdAccount.Parameters["@modified"].Value = account.Modified;
                        cmdAccount.Parameters["@icon"].Value = _encryptionService.EncryptString(account.Icon);
                        cmdAccount.Parameters["@notes"].Value = _encryptionService.EncryptString(account.Notes);
                        cmdAccount.Parameters["@autotype"].Value = _encryptionService.EncryptString(account.AutoTypeConfig.ToString());
                        cmdAccount.ExecuteNonQuery();

                        cmdAddFields.Parameters["@accountid"].Value = _encryptionService.EncryptString(account.ID);
                        foreach(IField field in account.Fields) {
                            try {
                                cmdAddFields.Parameters["@fieldname"].Value = _encryptionService.EncryptString(field.Name);
                                cmdAddFields.Parameters["@fieldvalue"].Value = _encryptionService.EncryptString(field.Value);
                                cmdAddFields.Parameters["@meta"].Value = _encryptionService.EncryptObject((int)field.Meta);
                                cmdAddFields.Parameters["@sortindex"].Value = field.SortIndex;
                                cmdAddFields.ExecuteNonQuery();
                            } catch(Exception ex) {
                                logger.Log(string.Format("Error adding field for account '{0}': {1}", account.Name, ex.Message), true);
                            }
                        }
                    } catch(Exception ex) {
                        logger.Log(string.Format("Error adding account: {0}", ex.Message), true);
                    }
                }
            }
        }

        /// <summary>
        /// Edits an account.
        /// </summary>
        /// <param name="account">The account.</param>
        public override void EditAccount(IAccount account) {
            SQLiteConnection conn = null;
            MultipleExceptionLogger logger = new MultipleExceptionLogger();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    using(SQLiteCommand cmd = conn.CreateCommand()) {
                        cmd.Transaction = trans;
                        cmd.CommandText = "UPDATE accounts SET name = @name, parentid = @parentid, templateid = @templateid, icon = @icon, notes = @notes, autotype = @autotype WHERE accountid = @accountid";
                        cmd.Parameters.AddWithValue("@accountid", _encryptionService.EncryptString(account.ID));
                        cmd.Parameters.AddWithValue("@name", _encryptionService.EncryptString(account.Name));
                        cmd.Parameters.AddWithValue("@parentid", _encryptionService.EncryptString(account.ParentID));
                        cmd.Parameters.AddWithValue("@templateid", _encryptionService.EncryptString(account.Template.ID));
                        cmd.Parameters.AddWithValue("@icon", _encryptionService.EncryptString(account.Icon));
                        cmd.Parameters.AddWithValue("@notes", _encryptionService.EncryptString(account.Notes));
                        cmd.Parameters.AddWithValue("@autotype", _encryptionService.EncryptString(account.AutoTypeConfig.ToString()));
                        cmd.ExecuteNonQuery();

                        cmd.CommandText = "DELETE FROM accountfields WHERE accountid = @accountid";
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "INSERT INTO accountfields (accountid, fieldname, fieldvalue, meta, sortindex) VALUES (@accountid, @fieldname, @fieldvalue, @meta, @sortindex)";
                        cmd.Parameters.Add("@fieldname", DbType.Binary);
                        cmd.Parameters.Add("@fieldvalue", DbType.Binary);
                        cmd.Parameters.Add("@meta", DbType.Binary);
                        cmd.Parameters.Add("@sortindex", DbType.Int32);
                        foreach(IField field in account.Fields) {
                            try {
                                cmd.Parameters["@fieldname"].Value = _encryptionService.EncryptString(field.Name);
                                cmd.Parameters["@fieldvalue"].Value = _encryptionService.EncryptString(field.Value);
                                cmd.Parameters["@meta"].Value = _encryptionService.EncryptObject((int)field.Meta);
                                cmd.Parameters["@sortindex"].Value = field.SortIndex;
                                cmd.ExecuteNonQuery();
                            } catch(Exception ex) {
                                logger.Log(string.Format("Error adding field for account '{0}': {1}", account.Name, ex.Message), true);
                            }
                        }
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
            if(logger.MessageLog.Count > 0) {
                logger.MessageLog.Insert(0, new Exception("Account was saved with errors"));
                throw new MultipleException(logger.MessageLog);
            }
        }

        /// <summary>
        /// Deletes an account.
        /// </summary>
        /// <param name="accountId">The account ID.</param>
        public override void DeleteAccount(string accountId) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                    using(SQLiteCommand cmd = conn.CreateCommand()) {
                        cmd.CommandText = "DELETE FROM accounts WHERE accountid = @accountid";
                        cmd.Parameters.AddWithValue("@accountid", _encryptionService.EncryptString(accountId));
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "DELETE FROM accountfields WHERE accountid = @accountid";
                        cmd.ExecuteNonQuery();
                    }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
        }

        /// <summary>
        /// Deletes multiple accounts.
        /// </summary>
        /// <param name="accountIds">The account IDs.</param>
        public override void DeleteAccounts(IEnumerable<string> accountIds) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = "DELETE FROM accounts WHERE accountid = @accountid";
                    cmd.Parameters.Add("@accountid", DbType.Binary);
                    foreach(var id in accountIds) {
                        cmd.Parameters["@accountid"].Value = _encryptionService.EncryptString(id);
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "DELETE FROM accountfields WHERE accountid = @accountid";
                        cmd.ExecuteNonQuery();
                    }
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
        }

        /// <summary>
        /// Gets a list of recipes from the current database file.
        /// </summary>
        /// <returns>A list of recipes in the database.</returns>
        public override List<IAccount> GetAccounts() {
            SQLiteConnection conn = null;
            ConsoleLogger logger = new ConsoleLogger();
            List<IAccount> accounts = new List<IAccount>();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                Dictionary<string, ITemplate> templates = new Dictionary<string,ITemplate>();
                using(SQLiteCommand cmdAccounts = conn.CreateCommand())
                using(SQLiteCommand cmdAccountFields = conn.CreateCommand())
                using(SQLiteCommand cmdTemplates = conn.CreateCommand()) {
                    cmdAccounts.CommandText = "SELECT accountid, name, parentid, templateid, created, modified, icon, notes, autotype FROM accounts";
                    cmdAccountFields.CommandText = "SELECT fieldname, fieldvalue, meta, sortindex FROM accountfields WHERE accountid = @accountid";
                    cmdAccountFields.Parameters.Add("@accountid", DbType.Binary);
                    cmdTemplates.CommandText = "SELECT name, shownotes FROM templates WHERE templateid = @templateid";
                    cmdTemplates.Parameters.Add("@templateid", DbType.Binary);
                    SQLiteDataReader reader = cmdAccounts.ExecuteReader();
                    while(reader.Read()) {
                        try {
                            AccountBase account = new AccountBase() {
                                ID = _encryptionService.DecryptString(GetBytes(0, reader)),
                                Name = _encryptionService.DecryptString(GetBytes(1, reader)),
                                ParentID = _encryptionService.DecryptString(GetBytes(2, reader)),
                                Created = Convert.ToDateTime(reader["created"]),
                                Modified = Convert.ToDateTime(reader["modified"]),
                                Icon = _encryptionService.DecryptString(GetBytes(6, reader)),
                                Notes = _encryptionService.DecryptString(GetBytes(7, reader)),
                                AutoTypeConfig = new LockCrypt.Core.AutoType.AutoTypeConfiguration(_encryptionService.DecryptString(GetBytes(8, reader))),
                                Template = new TemplateBase() {
                                    ID = _encryptionService.DecryptString(GetBytes(3, reader)),
                                }
                            };
                            accounts.Add(account);
                        } catch(Exception ex) {
                            logger.Log(string.Format("Error loading account: {0}", ex.Message), true);
                        }
                    }
                    reader.Close();
                    foreach(IAccount account in accounts) {
                        try {
                            cmdAccountFields.Parameters["@accountid"].Value = _encryptionService.EncryptString(account.ID);
                            reader = cmdAccountFields.ExecuteReader();
                            while(reader.Read()) {
                                string fieldMeta = _encryptionService.DecryptString(GetBytes(2, reader));
                                FieldBase field = new FieldBase() {
                                    Name = _encryptionService.DecryptString(GetBytes(0, reader)),
                                    Value = _encryptionService.DecryptString(GetBytes(1, reader)),
                                    Meta = string.IsNullOrEmpty(fieldMeta) ? FieldMetaData.None : (FieldMetaData)Enum.Parse(typeof(FieldMetaData), fieldMeta),
                                    SortIndex = Convert.ToInt32(reader["sortindex"])
                                };
                                account.Fields.Add(field);
                            }
                            reader.Close();
                            if(string.IsNullOrEmpty(account.Template.ID)) {
                                account.Template = new TemplateBase() { Name = "(Deleted)", ShowNotes = true };
                            } else {
                                if(templates.ContainsKey(account.Template.ID)) {
                                    account.Template = templates[account.Template.ID];
                                } else {
                                    cmdTemplates.Parameters["@templateid"].Value = _encryptionService.EncryptString(account.Template.ID);
                                    reader = cmdTemplates.ExecuteReader();
                                    if(reader.Read()) {
                                        TemplateBase template = new TemplateBase() {
                                            ID = account.Template.ID,
                                            Name = _encryptionService.DecryptString(GetBytes(0, reader)),
                                            ShowNotes = Convert.ToInt32(reader["shownotes"]) == 1
                                        };
                                        account.Template = template;
                                        templates.Add(template.ID, template);
                                    }
                                }
                            }
                        } catch(Exception ex) {
                            logger.Log(string.Format("Error loading account: {0}", ex.Message), true);
                        } finally {
                            reader.Close();
                        }
                    }
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }

            return accounts;
        }
        #endregion

        #region Groups
        /// <summary>
        /// Adds account groups to the default database.
        /// </summary>
        /// <param name="groups">The groups to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        public override void AddGroups(IEnumerable<IGroup> groups, ILogProvider logger) {
            SQLiteConnection conn = null;
            if(logger == null) {
                // if there's no logger, use the default.
                logger = _logger;
            }
            MultipleExceptionLogger logMultiple = new MultipleExceptionLogger();
            CompositeLogger logComposite = new CompositeLogger(logger, logMultiple);
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    AddGroups(conn, trans, groups, logComposite);
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
            if(logMultiple.MessageLog.Count > 0) {
                logMultiple.MessageLog.Insert(0, new Exception(I18nUtils.GetString("Errors", "NonFatalErrors")));
                throw new MultipleException(logMultiple.MessageLog);
            }
        }

        /// <summary>
        /// Adds groups using an existing transaction.
        /// </summary>
        /// <param name="connection">The connection.</param>
        /// <param name="transaction">The transaction.</param>
        /// <param name="groups">The groups.</param>
        /// <param name="logger">The logger.</param>
        private void AddGroups(SQLiteConnection connection, SQLiteTransaction transaction, IEnumerable<IGroup> groups, ILogProvider logger) {
            using(SQLiteCommand cmd = connection.CreateCommand()) {
                cmd.Transaction = transaction;
                cmd.CommandText = "INSERT INTO groups (groupid, name, parentid, sortindex) VALUES (@groupid, @name, @parentid, @sortindex)";
                cmd.Parameters.Add("@groupid", DbType.Binary);
                cmd.Parameters.Add("@name", DbType.Binary);
                cmd.Parameters.Add("@parentid", DbType.Binary);
                cmd.Parameters.Add("@sortindex", DbType.Int32);
                foreach(IGroup group in groups) {
                    try {
                        cmd.Parameters["@groupid"].Value = _encryptionService.EncryptString(group.ID);
                        cmd.Parameters["@name"].Value = _encryptionService.EncryptString(group.Name);
                        cmd.Parameters["@parentid"].Value = _encryptionService.EncryptString(group.ParentID);
                        cmd.Parameters["@sortindex"].Value = group.SortIndex;
                        cmd.ExecuteNonQuery();
                    } catch(Exception ex) {
                        logger.Log(string.Format("Error adding group: {0}", ex.Message), true);
                    }
                }
            }
        }

        /// <summary>
        /// Edits a group.
        /// </summary>
        /// <param name="group">The group.</param>
        public override void EditGroup(IGroup group) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    using(SQLiteCommand cmd = conn.CreateCommand()) {
                        cmd.CommandText = "UPDATE groups SET name = @name, parentid = @parentid, sortindex = @sortindex WHERE groupid = @groupid";
                        cmd.Parameters.AddWithValue("@groupid", _encryptionService.EncryptString(group.ID));
                        cmd.Parameters.AddWithValue("@name", _encryptionService.EncryptString(group.Name));
                        cmd.Parameters.AddWithValue("@parentid", _encryptionService.EncryptString(group.ParentID));
                        cmd.Parameters.AddWithValue("@sortindex", group.SortIndex);
                        cmd.ExecuteNonQuery();
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
        }

        /// <summary>
        /// Deletes a group and all child groups/accounts.
        /// </summary>
        /// <param name="group">The group.</param>
        public override void DeleteGroup(IGroup group) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    var groups = GetGroups();
                    IEnumerable<IGroup> nestedGroups = LockCryptBLL.ArrangeGroups(groups);
                    IGroup groupToDelete = null; // find the group in the heirarchy
                    foreach(var nestedGroup in nestedGroups) {
                        // search top level groups
                        groupToDelete = nestedGroup.FindGroup(group.ID);
                        if(groupToDelete != null)
                            break;
                    }
                    IEnumerable<IGroup> unavailableGroups = groupToDelete.Flatten();
                    IEnumerable<string> groupIds = unavailableGroups.Select(g => g.ID);

                    using(SQLiteCommand cmdDeleteGroups = conn.CreateCommand())
                    using(SQLiteCommand cmdDeleteAccounts = conn.CreateCommand()) {
                        cmdDeleteGroups.Transaction = trans;
                        cmdDeleteAccounts.Transaction = trans;
                        cmdDeleteGroups.CommandText = "DELETE FROM groups WHERE groupid = @groupid";
                        cmdDeleteGroups.Parameters.Add("@groupid", DbType.Binary);
                        cmdDeleteAccounts.CommandText = "DELETE FROM accounts WHERE parentid = @parentid";
                        cmdDeleteAccounts.Parameters.Add("@parentid", DbType.Binary);
                        foreach(string groupId in groupIds) {
                            cmdDeleteGroups.Parameters["@groupid"].Value = _encryptionService.EncryptString(groupId);
                            cmdDeleteGroups.ExecuteNonQuery();
                            cmdDeleteAccounts.Parameters["@parentid"].Value = _encryptionService.EncryptString(groupId);
                            cmdDeleteAccounts.ExecuteNonQuery();
                        }
                        //string inClause = string.Format("IN ('{0}')", groupIds.Aggregate((ids, id) => ids + "', '" + ));
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
        }

        /// <summary>
        /// Deletes multiple groups and their child accounts. Does not delete child groups.
        /// </summary>
        /// <param name="groupIds">The group IDs.</param>
        public override void DeleteGroups(IEnumerable<string> groupIds) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {

                    using(SQLiteCommand cmdDeleteGroups = conn.CreateCommand())
                    using(SQLiteCommand cmdDeleteAccounts = conn.CreateCommand()) {
                        cmdDeleteGroups.Transaction = trans;
                        cmdDeleteAccounts.Transaction = trans;
                        cmdDeleteGroups.CommandText = "DELETE FROM groups WHERE groupid = @groupid";
                        cmdDeleteGroups.Parameters.Add("@groupid", DbType.Binary);
                        cmdDeleteAccounts.CommandText = "DELETE FROM accounts WHERE parentid = @parentid";
                        cmdDeleteAccounts.Parameters.Add("@parentid", DbType.Binary);
                        foreach(string groupId in groupIds) {
                            cmdDeleteGroups.Parameters["@groupid"].Value = _encryptionService.EncryptString(groupId);
                            cmdDeleteGroups.ExecuteNonQuery();
                            cmdDeleteAccounts.Parameters["@parentid"].Value = _encryptionService.EncryptString(groupId);
                            cmdDeleteAccounts.ExecuteNonQuery();
                        }
                        //string inClause = string.Format("IN ('{0}')", groupIds.Aggregate((ids, id) => ids + "', '" + ));
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
        }

        /// <summary>
        /// Gets a list of account groups from the current database.
        /// </summary>
        /// <returns>A list of account groups in the database.</returns>
        public override List<IGroup> GetGroups() {
            SQLiteConnection conn = null;
            ConsoleLogger logger = new ConsoleLogger();
            List<IGroup> groups = new List<IGroup>();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = "SELECT groupid, name, parentid, sortindex FROM groups";
                    SQLiteDataReader reader = cmd.ExecuteReader();
                    while(reader.Read()) {
                        try {
                            GroupBase group = new GroupBase() {
                                ID = _encryptionService.DecryptString(GetBytes(0, reader)),
                                Name = _encryptionService.DecryptString(GetBytes(1, reader)),
                                ParentID = _encryptionService.DecryptString(GetBytes(2, reader)),
                                SortIndex = Convert.ToInt32(reader["sortindex"])
                            };
                            groups.Add(group);
                        } catch(Exception ex) {
                            logger.Log(string.Format("Error loading group: {0}", ex.Message), true);
                        }
                    }
                    reader.Close();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }

            return groups;
        }
        #endregion
        
        #region Templates
        /// <summary>
        /// Adds templates to the default database.
        /// </summary>
        /// <param name="templates">The templates to add.</param>
        /// <param name="logger">The delegate to use to log progress (can be null).</param>
        public override void AddTemplates(IEnumerable<ITemplate> templates, ILogProvider logger) {
            SQLiteConnection conn = null;
            List<string> errors = new List<string>();
            if(logger == null) {
                // if there's no log delegate, log to console.
                logger = new ConsoleLogger();
            }
            MultipleExceptionLogger logMultiple = new MultipleExceptionLogger();
            CompositeLogger logComposite = new CompositeLogger(logger, logMultiple);
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    AddTemplates(conn, trans, templates, logComposite);
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
            if(logMultiple.MessageLog.Count > 0) {
                logMultiple.MessageLog.Insert(0, new Exception(I18nUtils.GetString("Errors", "NonFatalErrors")));
                throw new MultipleException(logMultiple.MessageLog);
            }
        }

        /// <summary>
        /// Adds templates using an existing transaction.
        /// </summary>
        /// <param name="connection">The connection.</param>
        /// <param name="transaction">The transaction.</param>
        /// <param name="templates">The templates.</param>
        /// <param name="logger">The logger.</param>
        private void AddTemplates(SQLiteConnection connection, SQLiteTransaction transaction, IEnumerable<ITemplate> templates, ILogProvider logger) {
            using(SQLiteCommand cmdTemplate = connection.CreateCommand())
            using(SQLiteCommand cmdAddFields = connection.CreateCommand()) {
                cmdAddFields.Transaction = transaction;
                cmdTemplate.Transaction = transaction;
                cmdTemplate.CommandText = "INSERT INTO templates (templateid, name, shownotes, canuseredit) VALUES (@templateid, @name, @shownotes, @canuseredit)";
                cmdTemplate.Parameters.Add("@templateid", DbType.Binary);
                cmdTemplate.Parameters.Add("@name", DbType.Binary);
                cmdTemplate.Parameters.Add("@shownotes", DbType.Int32);
                cmdTemplate.Parameters.Add("@canuseredit", DbType.Int32);
                cmdAddFields.CommandText = "INSERT INTO templatefields (templateid, fieldname, meta, sortindex) VALUES (@templateid, @fieldname, @meta, @sortindex)";
                cmdAddFields.Parameters.Add("@templateid", DbType.Binary);
                cmdAddFields.Parameters.Add("@fieldname", DbType.Binary);
                cmdAddFields.Parameters.Add("@meta", DbType.Binary);
                cmdAddFields.Parameters.Add("@sortindex", DbType.Int32);
                foreach(ITemplate template in templates) {
                    try {
                        cmdTemplate.Parameters["@templateid"].Value = _encryptionService.EncryptString(template.ID);
                        cmdTemplate.Parameters["@name"].Value = _encryptionService.EncryptString(template.Name);
                        cmdTemplate.Parameters["@shownotes"].Value = template.ShowNotes ? 1 : 0;
                        cmdTemplate.Parameters["@canuseredit"].Value = template.CanUserEdit ? 1 : 0;
                        cmdTemplate.ExecuteNonQuery();

                        cmdAddFields.Parameters["@templateid"].Value = _encryptionService.EncryptString(template.ID);
                        foreach(IField field in template.Fields) {
                            try {
                                cmdAddFields.Parameters["@fieldname"].Value = _encryptionService.EncryptString(field.Name);
                                cmdAddFields.Parameters["@meta"].Value = _encryptionService.EncryptObject((int)field.Meta);
                                cmdAddFields.Parameters["@sortindex"].Value = field.SortIndex;
                                cmdAddFields.ExecuteNonQuery();
                            } catch(Exception ex) {
                                logger.Log(string.Format("Error adding field for template '{0}': {1}", template.Name, ex.Message), true);
                            }
                        }
                    } catch(Exception ex) {
                        logger.Log(string.Format("Error adding template: {0}", ex.Message), true);
                    }
                }
            }
        }

        /// <summary>
        /// Edits a template.
        /// </summary>
        /// <param name="template">The template.</param>
        public override void EditTemplate(ITemplate template) {
            SQLiteConnection conn = null;
            MultipleExceptionLogger logger = new MultipleExceptionLogger();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    using(SQLiteCommand cmd = conn.CreateCommand()) {
                        cmd.Transaction = trans;
                        cmd.CommandText = "UPDATE templates SET name = @name, shownotes = @shownotes WHERE templateid = @templateid";
                        cmd.Parameters.AddWithValue("@name", _encryptionService.EncryptString(template.Name));
                        cmd.Parameters.AddWithValue("@shownotes", template.ShowNotes ? 1 : 0);
                        cmd.Parameters.AddWithValue("@templateid", _encryptionService.EncryptString(template.ID));
                        cmd.ExecuteNonQuery();

                        cmd.CommandText = "DELETE FROM templatefields WHERE templateid = @templateid";
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "INSERT INTO templatefields (templateid, fieldname, meta, sortindex) VALUES (@templateid, @fieldname, @meta, @sortindex)";
                        cmd.Parameters.Add("@fieldname", DbType.Binary);
                        cmd.Parameters.Add("@meta", DbType.Binary);
                        cmd.Parameters.Add("@sortindex", DbType.Int32);
                        foreach(IField field in template.Fields) {
                            try {
                                cmd.Parameters["@fieldname"].Value = _encryptionService.EncryptString(field.Name);
                                cmd.Parameters["@meta"].Value = _encryptionService.EncryptObject((int)field.Meta);
                                cmd.Parameters["@sortindex"].Value = field.SortIndex;
                                cmd.ExecuteNonQuery();
                            } catch(Exception ex) {
                                logger.Log(string.Format("Error adding field for template '{0}': {1}", template.Name, ex.Message), true);
                            }
                        }
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
            if(logger.MessageLog.Count > 0) {
                logger.MessageLog.Insert(0, new Exception("Template was saved with errors"));
                throw new MultipleException(logger.MessageLog);
            }
        }

        /// <summary>
        /// Deletes a template.
        /// </summary>
        /// <param name="templateId">The template ID.</param>
        public override void DeleteTemplate(string templateId) {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = "DELETE FROM templates WHERE templateid = @templateid";
                    cmd.Parameters.AddWithValue("@templateid", _encryptionService.EncryptString(templateId));
                    cmd.ExecuteNonQuery();
                    cmd.CommandText = "DELETE FROM templatefields WHERE templateid = @templateid";
                    cmd.ExecuteNonQuery();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
        }

        /// <summary>
        /// Gets a list of templates from the current database file.
        /// </summary>
        /// <returns>A list of templates in the database.</returns>
        public override List<ITemplate> GetTemplates() {
            return GetTemplatesByName(null);
        }

        /// <summary>
        /// Gets matching templates by name.
        /// </summary>
        /// <param name="templateName">Name of the template.</param>
        /// <returns>All templates with a matching name.</returns>
        public override List<ITemplate> GetTemplatesByName(string templateName) {
            SQLiteConnection conn = null;
            ConsoleLogger logger = new ConsoleLogger();
            List<ITemplate> templates = new List<ITemplate>();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmdTemplates = conn.CreateCommand())
                using(SQLiteCommand cmdFields = conn.CreateCommand()) {
                    cmdTemplates.CommandText = "SELECT templateid, name, shownotes, canuseredit FROM templates";
                    if(!string.IsNullOrEmpty(templateName)) {
                        cmdTemplates.CommandText += " WHERE name = @name";
                        cmdTemplates.Parameters.AddWithValue("@name", _encryptionService.EncryptString(templateName));
                    }
                    cmdFields.CommandText = "SELECT fieldname, meta, sortindex FROM templatefields WHERE templateid = @templateid";
                    cmdFields.Parameters.Add("@templateid", DbType.Binary);
                    SQLiteDataReader reader = cmdTemplates.ExecuteReader();
                    while(reader.Read()) {
                        try {
                            TemplateBase template = new TemplateBase() {
                                ID = _encryptionService.DecryptString(GetBytes(0, reader)),
                                Name = _encryptionService.DecryptString(GetBytes(1, reader)),
                                ShowNotes = Convert.ToInt32(reader["shownotes"]) == 1,
                                CanUserEdit = Convert.ToInt32(reader["canuseredit"]) == 1
                            };
                            templates.Add(template);
                        } catch(Exception ex) {
                            logger.Log(string.Format("Error loading template: {0}", ex.Message), true);
                        }
                    }
                    reader.Close();
                    foreach(ITemplate template in templates) {
                        try {
                            cmdFields.Parameters["@templateid"].Value = _encryptionService.EncryptString(template.ID);
                            reader = cmdFields.ExecuteReader();
                            while(reader.Read()) {
                                string fieldMeta = _encryptionService.DecryptString(GetBytes(1, reader));
                                FieldBase field = new FieldBase() {
                                    Name = _encryptionService.DecryptString(GetBytes(0, reader)),
                                    Meta = string.IsNullOrEmpty(fieldMeta) ? FieldMetaData.None : (FieldMetaData)Enum.Parse(typeof(FieldMetaData), fieldMeta),
                                    SortIndex = Convert.ToInt32(reader["sortindex"])
                                };
                                template.Fields.Add(field);
                            }
                            reader.Close();
                        } catch(Exception ex) {
                            logger.Log(string.Format("Error loading fields for template '{0}': {1}", template.Name, ex.Message), true);
                        } finally {
                            reader.Close();
                        }
                    }
                }
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }

            return templates;
        }
        #endregion

        #region Maintenance
        /// <summary>
        /// Checks the password entered is valid.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if the current password is valid, otherwise <c>false</c>.
        /// </returns>
        public override bool CheckPassword() {
            return CheckPassword(false);
        }

        /// <summary>
        /// Checks the password entered is valid.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if the current password is valid, otherwise <c>false</c>.
        /// </returns>
        public bool CheckPassword(bool suppressExceptions) {
            bool valid = false;
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = @"SELECT value FROM meta WHERE name = @name;";
                    cmd.Parameters.AddWithValue("@name", _encryptionService.EncryptString("version"));
                    SQLiteDataReader reader = cmd.ExecuteReader();
                    if(reader.Read()) {
                        string versionString = _encryptionService.DecryptString(GetBytes(0, reader));
                        _dbVersion = new Version(versionString);
                        if(_dbVersion >= new Version("1.0.0.0")) {
                            valid = true;
                            _isConnected = true;
                        }
                    }

                }
            }catch(Exception ex) {
                if(!suppressExceptions) {
                    throw ex;
                }
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
            return valid;
        }

        /// <summary>
        /// Creates a database and required tables.
        /// </summary>
        /// <exception cref="System.Exception">Multiple exceptions, check the reference for DirectoryInfo.Create and the FileInfo constructor.</exception>
        public override void CreateDatabase() {
            FileInfo info = new FileInfo(_filename);
            if(!info.Directory.Exists) {
                info.Directory.Create();
            }
            SQLiteConnection.CreateFile(_filename);
            CreateTables();
        }

        /// <summary>
        /// Clears the database.
        /// </summary>
        public override void ClearDatabase() {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = @"DELETE FROM accounts;
                        DELETE FROM accountfields;
                        DELETE FROM templates;
                        DELETE FROM templatefields;
                        DELETE FROM groups;
                    ";
                    cmd.ExecuteNonQuery();
                }
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
        }

        /// <summary>
        /// Reads the whole contents of the database.
        /// </summary>
        /// <returns>The whole contents of the database.</returns>
        public override DatabaseContents ReadDatabase() {
            DatabaseContents contents = new DatabaseContents() {
                Accounts = GetAccounts(),
                Groups = GetGroups(),
                Templates = GetTemplates()
            };
            return contents;
        }

        /// <summary>
        /// Imports a database without clearing existing data.
        /// </summary>
        /// <param name="contents">The contents to import.</param>
        /// <param name="logger">The logger.</param>
        public override void ImportDatabase(DatabaseContents contents, ILogProvider logger) {
            AddAccounts(contents.Accounts, logger);
            AddGroups(contents.Groups, logger);
            AddTemplates(contents.Templates, logger);
        }

        /// <summary>
        /// Creates the database tables required in a specified file and adds the default account templates.
        /// </summary>
        public void CreateTables() {
            SQLiteConnection conn = null;
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteCommand cmd = conn.CreateCommand()) {
                    cmd.CommandText = @"CREATE TABLE accounts (
                            accountid BLOB PRIMARY KEY NOT NULL,
                            name BLOB NOT NULL,
                            parentid BLOB,
                            templateid BLOB,
                            created DATETIME NOT NULL,
                            modified DATETIME NOT NULL,
                            icon BLOB,
                            notes BLOB,
                            autotype BLOB
                        );
                        CREATE TRIGGER accountUpdateTime AFTER UPDATE ON accounts
                        BEGIN
                             UPDATE accounts SET modified = DATETIME('NOW') WHERE rowid = new.rowid;
                        END;
                        CREATE TRIGGER accountInsertTime AFTER INSERT ON accounts
                        BEGIN
                             UPDATE accounts SET created = DATETIME('NOW') WHERE rowid = new.rowid;
                             UPDATE accounts SET modified = DATETIME('NOW') WHERE rowid = new.rowid;
                        END;
                        CREATE TABLE templates (
                            templateid BLOB PRIMARY KEY NOT NULL,
                            name BLOB NOT NULL,
                            shownotes int NOT NULL DEFAULT 1,
                            canuseredit INT NOT NULL DEFAULT 1
                        );
                        CREATE TABLE templatefields (
                            templateid BLOB NOT NULL,
                            fieldname BLOB NOT NULL,
                            meta BLOB,
                            sortindex int NOT NULL DEFAULT 0,
                            PRIMARY KEY (templateid, fieldname)
                        );
                        CREATE TABLE accountfields (
                            accountid BLOB NOT NULL,
                            fieldname BLOB NOT NULL,
                            fieldvalue BLOB,
                            meta BLOB,
                            sortindex int NOT NULL DEFAULT 0,
                            PRIMARY KEY (accountid, fieldname)
                        );
                        CREATE TRIGGER fieldLastMod AFTER UPDATE on accountfields
                        BEGIN
                            UPDATE accounts SET modified = DATETIME('NOW') WHERE accountid = new.accountid;
                        END;
                        CREATE TABLE groups (
                            groupid BLOB PRIMARY KEY NOT NULL, 
                            name BLOB NOT NULL,
                            parentid BLOB, 
                            sortindex int NOT NULL DEFAULT 0
                        );
                        CREATE TABLE meta (
                            name BLOB PRIMARY KEY NOT NULL,
                            value BLOB
                        );
                    ";
                    cmd.ExecuteNonQuery();
                    cmd.CommandText = "INSERT INTO meta (name, value) VALUES (@name, @value);";
                    cmd.Parameters.Add("@name", DbType.Binary);
                    cmd.Parameters.Add("@value", DbType.Binary);
                    var values = new Dictionary<string, object>() {
                        { "version", Constants.GetVersion().ToString()},
                        { "platform", "WPF"}
                    };
                    foreach(KeyValuePair<string, object> value in values) {
                        cmd.Parameters["@name"].Value = _encryptionService.EncryptString(value.Key);
                        cmd.Parameters["@value"].Value = _encryptionService.EncryptObject(value.Value);
                        cmd.ExecuteNonQuery();
                    }
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch {
                    }
                }
            }
        }

        /// <summary>
        /// Changes the encryption password.
        /// </summary>
        /// <param name="newPassword">The new password.</param>
        public override void ChangePassword(string newPassword) {
            var db = ReadDatabase();
            SQLiteConnection conn = null;
            MultipleExceptionLogger logMultiple = new MultipleExceptionLogger();
            try {
                conn = GetConnection(_filename);
                conn.Open();
                using(SQLiteTransaction trans = conn.BeginTransaction()) {
                    List<KeyValuePair<string, string>> metaValues = new List<KeyValuePair<string, string>>();
                    using(SQLiteCommand cmd = conn.CreateCommand()) {
                        cmd.Transaction = trans;
                        cmd.CommandText = @"DELETE FROM accounts;
                                DELETE FROM accountfields;
                                DELETE FROM templates;
                                DELETE FROM templatefields;
                                DELETE FROM groups;";
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "SELECT name, value FROM meta";
                        SQLiteDataReader reader = cmd.ExecuteReader();
                        while(reader.Read()) {
                            metaValues.Add(new KeyValuePair<string, string>(_encryptionService.DecryptString(GetBytes(0, reader)), _encryptionService.DecryptString(GetBytes(1, reader))));
                        }
                        reader.Close();
                        Encryption.EncryptionKey = newPassword;
                        cmd.CommandText = "DELETE FROM meta";
                        cmd.ExecuteNonQuery();
                        cmd.CommandText = "INSERT INTO meta (name, value) VALUES (@name, @value);";
                        cmd.Parameters.Add("@name", DbType.Binary);
                        cmd.Parameters.Add("@value", DbType.Binary);
                        foreach(KeyValuePair<string, string> value in metaValues) {
                            cmd.Parameters["@name"].Value = _encryptionService.EncryptString(value.Key);
                            cmd.Parameters["@value"].Value = _encryptionService.EncryptString(value.Value);
                            cmd.ExecuteNonQuery();
                        }
                    }
                    try {
                        AddAccounts(conn, trans, db.Accounts, _logger);
                    } catch(Exception ex) {
                        logMultiple.MessageLog.Add(ex);
                    }
                    try {
                        AddGroups(conn, trans, db.Groups, _logger);
                    } catch(Exception ex) {
                        logMultiple.MessageLog.Add(ex);
                    }
                    try {
                        AddTemplates(conn, trans, db.Templates, _logger);
                    } catch(Exception ex) {
                        logMultiple.MessageLog.Add(ex);
                    }
                    trans.Commit();
                }
            } catch {
                throw;
            } finally {
                if(conn != null && conn.State == ConnectionState.Open) {
                    try {
                        conn.Close();
                    } catch { }
                }
            }
            if(logMultiple.MessageLog.Count > 0) {
                logMultiple.MessageLog.Insert(0, new Exception(I18nUtils.GetString("Errors", "NonFatalErrors")));
                throw new MultipleException(logMultiple.MessageLog);
            }
        }

        /// <summary>
        /// Creates, but does not open, a connection to a file.
        /// </summary>
        /// <param name="filename">The filename.</param>
        /// <returns>A connection to the file.</returns>
        private static SQLiteConnection GetConnection(string filename) {
            return new SQLiteConnection(string.Format(SQLiteConnectionString, filename));
        }

        /// <summary>
        /// Gets a byte array from a column.
        /// </summary>
        /// <param name="columnIndex">Index of the column.</param>
        /// <param name="reader">The reader.</param>
        /// <returns>The contents of a Blob from the column as a byte[].</returns>
        private static byte[] GetBytes(int columnIndex, IDataRecord reader) {
            if(reader.IsDBNull(columnIndex))
                return null;
            long length = reader.GetBytes(columnIndex, 0, null, 0, 0);
            byte[] myBinaryData = new byte[length];
            reader.GetBytes(columnIndex, 0, myBinaryData, 0, myBinaryData.Length);
            return myBinaryData;
        }

        #endregion
    }
}
